# -*- coding: utf-8 -*-
import collections
import operator

import colorama
import six


MAX_DEPTH = 10
TAB_WIDTH = 4
TAB = ' '
LIST_ITEM = '└─► '
COLOR_OUTPUT = True
BOLD = colorama.Style.BRIGHT
DIM = colorama.Style.DIM
NORMAL = colorama.Style.NORMAL
FG_BLACK = colorama.Fore.BLACK
FG_RED = colorama.Fore.RED
FG_GREEN = colorama.Fore.GREEN
FG_YELLOW = colorama.Fore.YELLOW
FG_BLUE = colorama.Fore.BLUE
FG_MAGENTA = colorama.Fore.MAGENTA
FG_CYAN = colorama.Fore.CYAN
FG_WHITE = colorama.Fore.WHITE
FG_NORMAL = colorama.Fore.RESET
BG_BLACK = colorama.Back.BLACK
BG_RED = colorama.Back.RED
BG_GREEN = colorama.Back.GREEN
BG_YELLOW = colorama.Back.YELLOW
BG_BLUE = colorama.Back.BLUE
BG_MAGENTA = colorama.Back.MAGENTA
BG_CYAN = colorama.Back.CYAN
BG_WHITE = colorama.Back.WHITE
BG_NORMAL = colorama.Back.RESET


def colored(fore, back, style):
    '''Formats the given colors and style into a string object'''
    return '%s%s%s' % (style, fore, back)


def formatter(list_item, color_output=COLOR_OUTPUT):
    '''Creates a string formatter for the given list item style'''
    color = '{color}' if color_output else ''
    color_reset = colorama.Style.RESET_ALL if color_output else ''
    return '{indent}%s%s{}%s\n' % (list_item, color, color_reset)


def get_attr(obj, attr):
    '''
    Gets the attribute value for the given attribute name and object.
    Allows nested attribute notation: attr.attr2
    :param obj: Object
    :param attr: String attribute name.
    :return value
    '''
    return operator.attrgetter(attr)(obj)


def rfilter(key, parent, sequence):
    '''
    :param key: callable, must accept two arguments, parent and child
    :param parent: Previous element from the sequence
    :param sequence: Iterable object
    '''
    if not key:
        return sequence
    return (c for c in sequence if key(parent, c))


def pprint(model, formattings, level=0):
    '''
    Pretty print the model with the given formatting parameters
    :param model: ``cosmicray.mode.Model`` object
    :param formattings: List of ``Formatting`` objects
    :param level: Indentation level. Default: 0

    Usage::
        class Dog(cosmicray.model.Model):
            __slots__ = ['id', 'name', 'breed']
            friends = cosmicray.model.relationship('Friend',
                urlargs={'id': cosmicray.model.ModelParam('id')}))
            owner = cosmicray.model.relationship('Owner',
                urlargs={'id': cosmicray.model.ModelParam('id')}))
            ...

        class Friend(cosmicray.model.Model):
            __slots__ = ['id', 'name', 'friendType', 'friendId']
            ...

        class Owner(cosmicray.model.Model):
            __slots__ = ['id', 'name']
            ...

        dog = Dog(id=12345).get()
        pprint(dog, formattings=[
            # Format header
            Formatting(formatter='{0.name} > {0.id}', color=FG_RED),
            # Format owners name
            Formatting('owner.name', color=FG_GREEN),
            # Format friends in hierarchical manner
            Formatting('friends',
                formatter='{0.name} > {0.id}',
                color=pprint.colored(pprint.FG_BLUE, pprint.BG_NORMAL, pprint.BOLD),
                is_recursive=True,
                is_sequence=True,
                rfilter=lambda parent, child: (
                    child.friendId == parent.id and child.friendType == (
                        'DOG' if isinstance(parent, Dog)
                        else 'FRIEND'))),
        ])
    '''
    colorama.init()
    pprinter = PrettyPrinter(TAB, TAB_WIDTH)
    pprinter.pprint(model, formattings, level)
    print(pprinter)


class Formatting(object):
    '''
    :param attr_name: A models attribute name. Optional.
    :param formatter: Formatting string using new style pythong formatting language.
        If ``attr_name`` is given formats the attributes value, otherwise formats the given model
    :param is_sequence: True|False. Indicates if attributes value is of type sequence. Default: False
    :param is_recursive: True|False. Used in conjunction with ``is_sequence`` and ``rfilter``.
        Indicates the given items in the sequence have a recursive relationship
    :param rfilter: Callable that must accept two arguments: parent and current, where parent
        is either a previous item from the sequence, or the model passed in to pprint
        Default: None
    :param color: Terminal color codes
    :param formattings: List of ``Formatting`` objects
    '''
    def __init__(self, attr_name=None, formatter=None, is_sequence=False,
                 is_recursive=False, rfilter=None, color=None, color_if=None,
                 formattings=None):
        self.attr_name = attr_name
        self.formatter = formatter
        self.is_sequence=is_sequence
        self.is_recursive = is_recursive
        self.rfilter = rfilter
        self.formattings = formattings or []
        self.color = color
        self.color_if = color_if

    def get_color(self, obj):
        if self.color_if:
            return self.color_if(obj)
        elif self.color:
            return self.color
        return ''


class PrettyPrinter(object):
    def __init__(self, tab, tab_width):
        self.writer = six.StringIO()
        self.formatter = formatter(list_item=LIST_ITEM)
        self.is_first = True
        self.tabs = tab * tab_width

    def write(self, obj):
        self.writer.write(obj)

    def writerow(self, obj, formatting, level, no_formatting=False):
        value = obj
        if formatting.formatter and not no_formatting:
            value = formatting.formatter.format(obj)
        color = formatting.get_color(value)
        if self.is_first and level == 0:
            self.write(formatter(list_item='').format(
                value, indent='', color=color))
            self.is_first = False
        else:
            self.write(
                self.formatter.format(
                    value, indent=(self.tabs * level), color=color))

    def __str__(self):
        return self.writer.getvalue()

    def rpprint(self, parent, sequence, formatting, level):
        if level >= MAX_DEPTH:
            return
        for item in rfilter(formatting.rfilter, parent, sequence):
            if formatting.formatter:
                self.writerow(item, formatting, level)
            if formatting.formattings:
                self.pprint(item, formatting.formattings, level + 1)
            if formatting.is_recursive:
                self.rpprint(item, sequence, formatting, level + 1)

    def pprint(self, model, formattings, level):
        if level >= MAX_DEPTH:
            return
        for formatting in formattings:
            increment = 0 if self.is_first else 1
            if formatting.attr_name:
                attr_value = get_attr(model, formatting.attr_name)
                if formatting.is_sequence:
                    self.writerow(formatting.attr_name, formatting, level, no_formatting=True)
                    self.rpprint(model, attr_value, formatting, level + increment)
                else:
                    self.writerow(attr_value, formatting, level)
            else:
                if formatting.is_sequence:
                    self.writerow(model.__class__.__name__,
                                  formatting, level, no_formatting=True)
                    self.rpprint(None, model, formatting, level)
                else:
                    self.writerow(model, formatting, level)
                    if formatting.formattings:
                        self.pprint(model, formatting.formattings, level)



# -*- coding: utf-8 -*-

import collections
import inspect
import json
import os
import string

import requests
import six

from . import util


STORE_ITEMS = ['auth', 'domain', 'headers', 'params', 'urlargs', 'extra']


class Cosmicray(object):
    '''
    :param name: App name and User-Agent header
    :param domain: Domain name. Default: http://localhost:8080
    :param home_dir: Apps home directory to store artifact files, such as credentials
        the full path for the home directory will be ``~/.cosmicray/{name}``

    Usage::

        >>> api = Cosmicray('myapp', 'http://mydomain.com')
        >>> api.configure(headers={'Content-Type': 'application/json'})

    '''
    def __init__(self, name, domain='http://localhost:8080', home_dir=None):
        self.name = name
        self.routes = []
        self.config = util.Config({
            'debug': False,
            'raise_for_status': True,
            'disable_validation': False,
            'home_dir': util.create_home_dir(name, root_path=home_dir)
        })
        self.config['config_filename'] = self.home_dir('config')
        self.tpl = util.RequestTemplate(
            domain=domain, headers={'User-Agent': self.name})
        self.session = requests.Session()

    def home_dir(self, *args):
        '''Returns home directory path joined with the given argument sequence'''
        return os.path.join(self.config['home_dir'], *args)

    def cache_dir(self, *args):
        path = self.home_dir('cache')
        if not os.path.exists(path):
            os.makedirs(path)
        return os.path.join(path, *args)

    def route(self, path, methods, params=None, urlargs=None, headers=None):
        '''
        Decorates a function that needs to accept :class:`requests.Response <Response>` object
        as the only argument, process the response and return the result (ex. return json).

        :param path: Uri of type str or string formatter.
        :param methods: list of all allowed methods
        :param params: list of :class:`cosmicray.Param` query parameters
        :param urlargs: list of :class:`cosmicray.Param`
        :param headers: mapping representing request headers
        :returns: decorator

        Usage::

            >>> api = Cosmicray('myapp', 'http://mydomain.com')
            >>> @api.route('/some/path')
            ... def some_path(response):
            ...     return response.json()
            ...

        '''
        route = Route(path, methods, params, urlargs, headers, app=self)
        self.routes.append(route)
        return route.response_handler_decorator

    def configure(self, config=None, **kwargs):
        '''
        :param config: Mapping for general app settings
        :param \*\*kwargs: Application wide request configurations
             domain: Domain name
             headers: Mapping of http headers
             urlargs: Mapping of url formatting arguments
             params: Mapping of query parameters
             authenticator: Callback for request authentication
             auth: Basic auth
             &rest: Keyword arguments for `requests`
        '''
        self.tpl.update(**kwargs)
        self.config.update(config)

    def get_config(self, key):
        '''Return config value for the given key'''
        try:
            return getattr(self.tpl, key.lower())
        except AttributeError:
            try:
                return self.config[key]
            except KeyError:
                try:
                    return self.tpl.extra[key]
                except KeyError:
                    pass

    def store_configurations(self):
        ''''Save config to file'''
        configs = {
            'config': self.config.dict,
            'request_config': dict(
                (k, v) for k, v in self.tpl.items() if k in STORE_ITEMS),
        }
        util.write_artifact_file(
            self.get_config('config_filename'), configs, json.dumps)

    def load_configurations(self):
        ''''Load config from file'''
        fpath = self.get_config('config_filename')
        if os.path.exists(fpath):
            configs = util.read_artifact_file(fpath, json.loads)
            self.config.update(configs.get('config'))
            self.tpl.update(**configs.get('request_config', {}))

    def __repr__(self):
        return '<{app} app for {name}>'.format(
            app=self.__class__.__name__, name=self.name)


class Route(object):
    '''Defines properties of a route

    :param path: Uri of type str or string formatter.
    :param methods: list of all allowed methods
    :param params: list of :class:`cosmicray.Param` query parameters
    :param urlargs: list of :class:`cosmicray.Param`
    :param headers: mapping representing request headers
    '''

    def __init__(self, path, methods, params, urlargs, headers, app=None):
        self.app = app
        self.path = path
        self.methods = methods
        self.params = params
        self.urlargs = urlargs
        self.headers = headers
        self.response_handler = None

    def response_handler_decorator(self, response_handler):
        '''Decorates a function and turns it into response handler'''
        self.response_handler = response_handler
        return self

    def __eq__(self, obj):
        if isinstance(obj, Route):
            return self.path == obj.path
        elif isinstance(obj, six.string_types):
            return self.path == obj
        raise TypeError('Incompatible type: {}'.format(type(obj)))

    def __call__(self, model_cls=None, **kwargs):
        request = Request(
            route=self, model_cls=model_cls,
            args=self.app.tpl.items(),
            path=self.path, headers=self.headers)
        request.update(**kwargs)
        return request

    def get_config(self, key):
        '''Return config value for the given key'''
        return self.app.get_config(key)

    def __repr__(self):
        return '<{name} for {path!r}>'.format(
            name=self.__class__.__name__, path=self.path)


class Request(util.RequestTemplate):
    __doc__ = ('''
    :param route: instance of :class:`Route`
    :param model_cls: class that implements `_make` method
    :param args: sequence of key-value pairs to initialize the request with
    :param \*\*kwargs: keyword arguments to initialize the request with

    Any parameters given to :class:`Request` can be updated with :class:`Request.update` OR
    completly overwritten with :class:`Request.override`

    :class:`Request` provides special setter methods that set the given attribute and return the instance of itself to chain together calls:
    {}
    '''.format('\n'.join('\t\t:class:`Request.set_{}`\n'.format(attr)
                       for attr in util.RequestTemplate.__slots__)))
    __slots__ = ()

    def __init__(self, route, model_cls, args=None, **kwargs):
        super(Request, self).__init__(args, **kwargs)
        self.route = route
        self.model_cls = model_cls

    def is_request_for(self, *routes):
        '''Returns True if :class:`Request`.route is in the given sequence of routes'''
        return any(self.route == route for route in routes)

    def handle_response(self, response):
        '''Calls the routes response handler with the given response and maps the model to the given result'''
        expected_args = inspect.getargspec(self.route.response_handler).args
        args = [self, response] if len(expected_args) == 2 else [response]
        return self.map_model(self.route.response_handler(*args))

    def map_model(self, response):
        '''Calls :class:`Request`.model_cls._make method if a model was provided, otherwise returns the given response'''
        if self.model_cls is not None:
            _make = getattr(self.model_cls, '_make')
            try:
                return _make(response)
            except TypeError:
                return map(_make, response)
        return response

    def authenticate(self):
        '''Authenticates the request, if :class:`Request`.authenticator call back is provided'''
        if self.authenticator:
            return self.authenticator(self)
        return self

    def validate(self):
        '''Validates method, query parameters, and urlarg parameters'''
        if not self.route.get_config('disable_validation'):
            if self.method not in self.route.methods:
                raise TypeError('Method {!r} is not supported by {!r}'.format(
                    self.method, self.route))

            self.params = self._validate_params(self.params, self.route.params)
            self.urlargs = self._validate_params(self.urlargs, self.route.urlargs)
        return self

    def _validate_params(self, actual, expected):
        if expected:
            for param in expected:
                value = param.validate(actual.get(param.name), context=self)
                actual[param.name] = value
        return actual

    @property
    def url(self):
        '''combines domain and path and formats the final result with urlargs'''
        uri = DefaultUrlFormatter().format(self.path, **self.urlargs)
        return '{}/{}'.format(self.domain.rstrip('/'), uri.lstrip('/'))

    def _log(self, response):
        '''Log the request and response'''
        if self.route.get_config('debug'):
            obj = json.dumps({
                'request': {
                    'url': response.url,
                    'method': response.request.method,
                    'headers': dict(response.request.headers),
                    'body': str(response.request.body or "")
                },
                'response': {
                    'headers': dict(response.headers),
                    'status_code': response.status_code,
                    'body': response.text,
                    'duration': response.elapsed.total_seconds()
                }
            })
            print(obj)

    def send(self):
        '''Makes request and returns the result of the response being handled by the response handler'''
        response = self.route.app.session.request(
            self.method.lower(), self.url, headers=self.headers, params=self.params,
            data=self.data, files=self.files, json=self.json,
            auth=self.auth, **self.extra)
        self._log(response)
        if self.route.get_config('raise_for_status'):
            response.raise_for_status()
        return self.handle_response(response)

    def get(self):
        '''GET request'''
        return self.set_method('GET').validate().authenticate().send()

    def delete(self):
        '''DELETE request'''
        return self.set_method('DELETE').validate().authenticate().send()

    def post(self):
        '''POST request'''
        return self.set_method('POST').validate().authenticate().send()

    def put(self):
        '''PUT request'''
        return self.set_method('PUT').validate().authenticate().send()

    def head(self):
        '''HEAD request'''
        return self.set_method('HEAD').validate().authenticate().send()

    def options(self):
        '''OPTIONS request'''
        return self.set_method('OPTIONS').validate().authenticate().send()

    def __repr__(self):
        return '<Request for {}>'.format(self.route.path)


class DefaultUrlFormatter(string.Formatter):

    def get_value(self, key, args, kwargs):
        try:
            return kwargs[key]
        except KeyError as error:
            return ''
        return Formatter.get_value(key, args, kwargs)


class Param(object):
    '''Used for query parameters and urlargs to validate and specify default values

    :param name: the parameters name
    :param default: Specify default value for the parameter. The default may be a
        callback that accepts two arguments, an object and context, where object
        is the given value for the parameter and context is :class:`Request`
    :param required: boolean to indicate if parameter is required
    :param options: Sequence of options that the parameter can have
    '''
    def __init__(self, name, default=None, required=False, options=None):
        self.name = name
        self.default = default
        self.required = required
        self.options = options

    def validate(self, obj, context):
        '''Validates the parameter and if it's invalid raises an exception

        :param obj: value provided for the parameter
        :param context: :class:`Request` instance
        :returns: value for the parameter
        '''
        if callable(self.default):
            value = self.default(obj, context)
        else:
            value = self.default if obj is None else obj

        if self.required and not value:
            raise TypeError('Required parameter {!r} is None or not provided'.format(self.name))

        if self.options and value not in self.options:
            raise TypeError('Invalid value for parameter {!r}: {!r}'.format(self.name, value))

        return value

    def __repr__(self):
        return '{name}({args})'.format(
            name=self.__class__.__name__, args=', '.join(
            ['{}={!r}'.format((attr, getattr(self, attr)))
             for attr in ['name', 'default', 'required', 'options']]))



# -*- coding: utf-8 -*-
import collections
import inspect
import importlib
import time


class Model(object):
    __slots__ = (
        '__changes__',
        '__model_attr__'
    )
    __ignore__ = []
    __route__ = None

    def __init__(self, **kwargs):
        self.__changes__ = []
        self.__model_attr__ = {}
        self.setfields(kwargs, track=False)
        self.clear_changes()

    def setfields(self, fields, track=True):
        for field in self.__slots__:
            try:
                default = getattr(self, field)
            except AttributeError:
                default = None
            value = fields.pop(field, default)
            setattr(self, field, value)
            if track:
                self.track_changes(field, value)
        for field in self.__ignore__:
            fields.pop(field, None)
        if fields:
            print('{!r} got extra fields: {}'.format(
                self.__class__.__name__, ', '.join(fields.keys())))

    def items(self):
        '''Returns iterator of key-value pairs from the objects fields'''
        return ((field, getattr(self, field)) for field in self.__slots__)

    def get_dict(self):
        '''Returns dict representation of the object'''
        return dict((k, v) for k, v in self.items())

    def set_dict(self, fields):
        '''Updates fields from the given dict object'''
        self.setfields(fields)

    dict = property(get_dict, set_dict, doc='Getter and setter from dict object')

    def __nonzero__(self):
        return any(value is not None for _, value in self.items())

    def track_changes(self, field, new_value):
        '''Tracks which fields have updated values'''
        self.__changes__.append(field)

    def clear_changes(self):
        '''Clears tracking of changes'''
        self.__changes__ = []

    def get_create_payload(self):
        '''POST request default payload: { "json" : self.dict }'''
        return {'json': self.dict}

    def get_update_payload(self):
        '''PUT request default payload: { "json" : self.dict }'''
        return {'json': self.dict}

    def __call__(self, **kwargs):
        '''Returns :class:`cosmicray.Request` with :class:`Model`.__class__ as the model_cls
        and ``self.dict`` as urlargs'''
        return self.__route__(
            model_cls=self.__class__, urlargs=self.dict, **kwargs)

    def get(self):
        '''GET request'''
        self.dict = self.__route__(urlargs=self.dict).get()
        return self

    def delete(self):
        '''DELETE request'''
        return self().delete()

    def create(self):
        '''POST request. Uses :class:`Model.get_create_payload` as the POST body'''
        return self(**self.get_create_payload()).post()

    def update(self):
        '''PUT request. Uses :class:`Model.get_update_payload` as the PUT body'''
        return self(**self.get_update_payload()).put()

    def update_related(self):
        '''Sends PUT request for all related models and routes'''
        for model_attr in self.__model_attr__.values():
            model_attr.update()

    def create_related(self):
        '''Sends POST request for all related models and routes'''
        for model_attr in self.__model_attr__.values():
            model_attr.create()

    def delete_related(self):
        '''Sends DELETE request for all related models and routes'''
        for model_attr in self.__model_attr__.values():
            model_attr.delete()

    @classmethod
    def _make(cls, fields):
        return cls(**fields)

    def __repr__(self):
        changed = ' has pending updates' if self.__changes__ else ''
        fields = ', '.join('{}={!r}'.format(f, v) for f, v in self.items())
        return '<{model}({fields}){changed}>'.format(
            model=self.__class__.__name__, fields=fields, changed=changed)


class ModelAttribute(object):
    '''
    :param model_cls_name: Class name for the related model. If both route and model class are provided, the
        route will take precedence.
    :param route: instance of ``cosmicray.Route``. If both route and model class are provided, the
        route will take precedence.
    :params urlargs: URL formatting parameters as dict object
    :param params: URL query parameters as dict object
    :param is_sequence: Specify True if attribute is a list of Models, False otherwise
    :param get_create_payload: Callable, accepts instance of ``cosmicray.Model`` as its only
        argument and returns POST request payload arguments
    :param get_update_payload: Callable, accepts instance of ``cosmicray.Model`` as its only
        argument and returns PUT request payload arguments
    :param get: Overrides ``GET`` request. Callable which accepts two arguments, ``model_ref`` and ``model_obj``, where ``model_ref`` is an instance of the class that the model attribute belongs to. And ``model_obj`` is an instance of ``ModelInstanceAttribute``, in other words, the related model object or response from the request
    :param create:  Overrides ``POST`` request. Callable which accepts two arguments, ``model_ref`` and ``model_obj``, where ``model_ref`` is an instance of the class that the model attribute belongs to. And ``model_obj`` is an instance of ``ModelInstanceAttribute``, in other words, the related model object or response from the request
    :param update: Overrides ``PUT`` request. Callable which accepts two arguments, ``model_ref`` and ``model_obj``, where ``model_ref`` is an instance of the class that the model attribute belongs to. And ``model_obj`` is an instance of ``ModelInstanceAttribute``, in other words, the related model object or response from the request
    :param delete: Overrides ``DELETE`` request. Callable which accepts two arguments, ``model_ref`` and ``model_obj``, where ``model_ref`` is an instance of the class that the model attribute belongs to. And ``model_obj`` is an instance of ``ModelInstanceAttribute``, in other words, the related model object or response from the request
    :param ttl: Timeout in seconds for how long to cache result of a request. Default: None, never expire
    :param is_static: Specify if attribute can be accessed like a class attribute

    '''
    def __init__(self, model_cls_name, route, urlargs, params, is_sequence,
                 get_create_payload, get_update_payload,
                 get=None, create=None, update=None, delete=None, ttl=None,
                 is_static=False, lazy=False):
        self.model_cls_name = model_cls_name
        self.route = route
        self.urlargs = urlargs or {}
        self.params = params or {}
        self.is_sequence = is_sequence
        self.get_update_payload = get_update_payload
        self.get_create_payload = get_create_payload
        self.get_method = get
        self.create_method = create
        self.update_method = update
        self.delete_method = delete
        self.ttl = ttl
        self.is_static = is_static
        self.lazy = lazy
        self._static = None

    def __get__(self, model_ref, model_cls):
        if self.is_static:
            return self.get_static_model_instance_attribute(model_cls).getter()
        return self._get_model_instance_attribute(model_ref).getter()

    def __set__(self, model_ref, value):
        if self.is_static:
            raise AttributeError('Cannot set static attribute')
        return self._get_model_instance_attribute(model_ref).setter(value)

    def __delete__(self, model_ref):
        if self.is_static:
            raise AttributeError('Cannot delete static attribute')
        return self._get_model_instance_attribute(model_ref).deleter()

    def _get_model_instance_attribute(self, model_ref):
        if self not in model_ref.__model_attr__:
            model_ref.__model_attr__[self] = ModelInstanceAttribute(
                model_attr=self, model_ref=model_ref)
        return model_ref.__model_attr__[self]

    def get_static_model_instance_attribute(self, model_cls):
        if self._static is None:
            self._static = ModelInstanceAttribute(model_attr=self, model_ref=model_cls)
        return self._static

    def __repr__(self):
        return '<{} for {}>'.format(self.__class__.__name__, self.model_cls_name)


class ModelInstanceAttribute(object):
    '''
    :param model_attr: Instance of ``cosmicray.model.ModelAttribute``
    :param model_ref: Instance of sub-class of``cosmicray.model.Model``
    '''
    def __init__(self, model_attr, model_ref):
        self.model_attr = model_attr
        self.model_ref = model_ref
        self.model_obj = None
        self._fetched_on = None

    @property
    def model_cls(self):
        if self.model_attr.model_cls_name is not None:
            module_name, _, classname = self.model_attr.model_cls_name.rpartition('.')
            module = importlib.import_module(module_name)
            return getattr(module, classname)

    @property
    def route(self):
        return self.model_attr.route

    @property
    def value(self):
        return self.model_obj

    def __nonzero__(self):
        return self.model_obj is not None

    def __getattr__(self, attr):
        return getattr(self.model_obj, attr)

    def __getitem__(self, name):
        return self.model_obj.__getitem__(name)

    def __setitem__(self, name, value):
        return self.model_obj.__setitem__(name, value)

    def __delitem__(self, name):
        return self.model_obj.__delitem__(name)

    def __iter__(self):
        return self.model_obj.__iter__()

    def __next__(self):
        return self.model_obj.__next__()

    def clear(self):
        self.model_obj = None

    def clear_if_expired(self):
        if self.expired():
            self.clear()

    def getter(self):
        if not self.model_attr.lazy:
            self.clear_if_expired()
            if self.model_obj is None:
                self.get()
        return self

    def setter(self, obj):
        self.model_obj = obj

    def deleter(self):
        self.clear()

    def __call__(self, **kwargs):
        urlargs = {}
        for name, param in self.model_attr.urlargs.items():
            if isinstance(param, ModelParam):
                urlargs[name] = param.validate(getattr(self.model_ref, param.name))
            else:
                urlargs[name] = param
        params = {}
        for name, param in self.model_attr.params.items():
            if isinstance(param, ModelParam):
                params[name] = param.validate(getattr(self.model_ref, param.name))
            else:
                params[name] = param

        if self.model_attr.route:
            request = self.model_attr.route(
                model_cls=self.model_cls, urlargs=urlargs, params=params)
        else:
            request = self.model_cls()()\
                          .set_urlargs(urlargs)\
                          .set_params(params)
        request.update(**kwargs)
        return request

    def expired(self):
        return (self.model_attr.ttl
                and self._fetched_on and (
                    (time.time() - self._fetched_on) > self.model_attr.ttl))

    def get(self):
        if self.model_attr.get_method:
            self.model_obj = self._as_sequence(
                self.model_attr.get_method(self.model_ref, self))
        else:
            self.model_obj = self._as_sequence(self().get())
        self._fetched_on = time.time()
        return self.model_obj

    def update(self):
        if self.model_attr.update_method:
            return self._as_sequence(
                self.model_attr.update_method(self.model_ref, self))
        if self.model_attr.get_update_payload is None:
            return self._as_sequence(self.model_obj.update())
        return self._as_sequence(
            self(**self.model_attr.get_update_payload(self.model_ref)).put())

    def create(self):
        if self.model_attr.create_method:
            return self._as_sequence(
                self.model_attr.create_method(self.model_ref, self))
        if self.model_attr.get_create_payload is None:
            return self._as_sequence(self.model_obj.create())
        return self._as_sequence(
            self(**self.model_attr.get_create_payload(self.model_ref)).post())

    def delete(self):
        if self.model_attr.delete_method:
            return self._as_sequence(
                self.model_attr.delete_method(self.model_ref, self))
        return self().delete()

    def _as_sequence(self, result):
        return list(result) if self.model_attr.is_sequence else result

    def __repr__(self):
        return repr(self.model_obj)

    def __str__(self):
        return str(self.model_obj)


def relationship(
        model_cls=None, route=None, urlargs=None, params=None,
        is_sequence=False,
        get_create_payload=None, get_update_payload=None,
        get=None, create=None, update=None, delete=None,
        ttl=None, is_static=False, lazy=False):
    '''Returns property-like object with a getter, setter, and deleter method'''
    if model_cls is None and route is None:
        raise ValueError("model_cls and route cannot both be None")
    frame = inspect.stack()[1]
    module_name = inspect.getmodule(frame[0]).__name__
    model_cls_name = '{}.{}'.format(module_name, model_cls)
    model_attr = ModelAttribute(
        model_cls_name, route, urlargs, params, is_sequence,
        get_create_payload, get_update_payload,
        get, create, update, delete, ttl, is_static, lazy)
    return model_attr


class ModelParam(object):
    def __init__(self, name):
        self.name = name

    def validate(self, obj, context=None):
        if obj is None:
            raise TypeError('Invalid value for {}: {!r}'.format(self.name, obj))
        return obj



# -*- coding: utf-8 -*-


import codecs
import collections
import copy
import functools
import os
import threading
import time


COSMICRAY_DIR = '~/.cosmicray'


class RequestTemplate(object):

    DEFAULTS = {
        'headers': collections.OrderedDict,
        'params': dict,
        'urlargs': dict,
        'extra': dict
    }
    __slots__ = (
        'auth',
        'data',
        'files',
        'headers',
        'json',
        'params',
        'extra',
        'authenticator',
        'domain',
        'path',
        'method',
        'urlargs',
        'model_cls',
        'route'
    )

    def __init__(self, args=None, **kwargs):
        for attr in RequestTemplate.__slots__:
            default = RequestTemplate.DEFAULTS.get(attr)
            setattr(self, attr, None if not default else default())
        self.update(args, **kwargs)

    def items(self):
        '''Returns iterator of attr and value pairs. The returned value is a copy'''
        return ((attr, copy.deepcopy(getattr(self, attr)))
                for attr in RequestTemplate.__slots__)

    def copy(self):
        '''Returns new instance of :class:`RequestTemplate` with a copy of all fields'''
        return self.__class__(self.items())

    def update(self, args=None, **kwargs):
        '''Soft updates based on the given arguments. Mapping objects will be
        merged rather than overwritten'''
        if args:
            self._update_from_sequence(args)
        self._update_from_map(kwargs)

    def override(self, args=None, **kwargs):
        '''Updates based on the given arguments. Mapping objects will be overwritten'''
        if args:
            self._update_from_sequence(args, override=True)
        self._update_from_map(kwargs, override=True)

    def _update_from_sequence(self, arg, override=False):
        for attr, value in arg:
            self._set_attr(attr, value, {}, override)

    def _update_from_map(self, kwargs, override=False):
        for attr, value in kwargs.items():
            self._set_attr(attr, value, {}, override)

    def _set_attr(self, attr, arg, kwargs, override):
        '''
        Sets the value for the given attribute
        :param attr: the attribute name
        :param arg: value or sequence
        :param kwargs: Mapping
        :param override: If the attribute being set is a mapping, then if override is True,
            it will discard existing values. Otherwise, it updates the existing mapping
        '''
        try:
            if not override:
                obj = getattr(self, attr)
                try:
                    obj.update(arg or {}, **kwargs)
                    arg = obj
                except AttributeError:
                    pass
            setattr(self, attr, arg)
        except AttributeError:
            self.extra[attr] = arg

    def _set_attr_chain(self, attr):
        '''All attributes in the :class:`RequestTemplate`.__slots__ have a
        special setter that can be used to set the value of the attribute and
        return the instance of itself to chain together calls.
        The setter is of the form: set_{attr name}
        '''
        def wrapper(arg=None, override=False, **kwargs):
            self._set_attr(attr, arg, kwargs, override=override)
            return self
        return wrapper

    def __getattribute__(self, attr):
        if attr.startswith('set_'):
            _, attr = attr.split('set_')
            if attr in RequestTemplate.__slots__:
                return self._set_attr_chain(attr)
        elif attr in ['urlargs', 'params']:
            # Filter out none values from urlargs
            obj = object.__getattribute__(self, attr)
            if obj:
                return dict((k, v) for k, v in obj.items() if v is not None)
            return obj
        return object.__getattribute__(self, attr)


class CachedArtifact(object):
    '''
    Provides simple, thread-safe, decorators to cache frequently accessed files
    and update the cache automatically if files get written to.
    '''

    __cached = {}
    _lock = threading.Lock()

    @classmethod
    def read(cls, function):
        '''Returns cached value for the given path, otherwise reads from disk'''
        @functools.wraps(function)
        def decorate(fpath, serializer=None):
            while CachedArtifact._lock.locked():
                time.sleep(0.001)
            if not fpath in cls.__cached:
                cls.__cached[fpath] = function(fpath, serializer)
            return cls.__cached[fpath]
        return decorate

    @classmethod
    def write(cls, function):
        '''Writes data for the given path to disk and updates the cache'''
        @functools.wraps(function)
        def decorate(fpath, data, serializer=None):
            with CachedArtifact._lock:
                try:
                    del cls.__cached[fpath]
                except KeyError:
                    pass
                cls.__cached[fpath] = function(fpath, data, serializer)
                return cls.__cached[fpath]
        return decorate


def _keys_to_lower(obj):
    return dict((k.lower(), v) for k, v in obj.items())


class Config(object):
    '''lettercase-agnostic key-value store'''
    def __init__(self, args=None,  **kwargs):
        self._config = {}
        self.update(args, **kwargs)

    def get(self, key, default=None):
        '''Return value for the given key or default if key not found'''
        return self._config.get(key.lower(), default)

    def setdefault(self, key, default):
        '''Set default value for a key if it doesn't exist'''
        return self._config.setdefault(key.lower(), default)

    def keys(self):
        return self._config.keys()

    def getcopy(self, key, default=None):
        '''Returns a copy of the value for the given key or default if key not found'''
        return copy.deepcopy(self.get(key, default))

    def __getitem__(self, key):
        return self._config[key.lower()]

    def __setitem__(self, key, value):
        self._config[key.lower()] = value

    def __delitem__(self, key):
        del self._config[key.lower()]

    def update(self, args=None, **kwargs):
        '''Updates key-value store'''
        self._config.update(
            _keys_to_lower(args or {}), **_keys_to_lower(kwargs))

    def copy(self):
        '''Return instance of :class:`Config` with copy of the objects data'''
        return Config(copy.deepcopy(self._config))

    def __repr__(self):
        return '<Config {!r}>'.format(self._config)

    def get_dict(self):
        return self._config

    def set_dict(self, obj):
        self._config = obj

    def del_dict(self):
        del self._config

    dict = property(get_dict, set_dict, del_dict)


def create_home_dir(name=None, root_path=None):
    '''Creates a home directory

    :param name: Optional directory name.
    :param root_path: Alternative root path. Default: ~/.cosmicray
    :returns: home directory path
    '''
    if root_path is None:
        root_path = os.path.expanduser(COSMICRAY_DIR)
    directory = os.path.join(root_path, name or '')
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory


@CachedArtifact.write
def write_artifact_file(fpath, data, serializer=None):
    '''Writes serialized data to the given file path'''
    with codecs.open(fpath, 'w', 'utf-8') as fobj:
        fobj.write(serializer(data) if serializer else data)
    return data


@CachedArtifact.read
def read_artifact_file(fpath, serializer=None):
    '''Reads and deserializes contents of data to the given file path'''
    with codecs.open(fpath, 'r', 'utf-8') as fobj:
        data = fobj.read()
        return serializer(data) if serializer else data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

AUTHOR = 'Samir Omerovic'
EMAIL = 'somerovi@gmail.com'
URL = 'https://github.com/somerovi/cosmicray'
NAME = 'cosmicray'
SUMMARY = "Develop a client for your HTTP API and document its quirks and features"
DESCRIPTION = ''
PACKAGES = ['cosmicray']
REQUIRED = [
    'colorama==0.3.9',
    'requests==2.18.4',
    'six==1.11.0'
]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    DESCRIPTION = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)

class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        cmd = '{0} setup.py sdist bdist_wheel --universal'.format(sys.executable)
        os.system(cmd)

        twine = sys.executable.replace('python', 'twine')
        self.status('Uploading the package to PyPi via Twine…')
        os.system('{0} upload dist/*'.format(twine))

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=SUMMARY,
    long_description=DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=PACKAGES,
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)
